function [ phi, sig2 ] = ARestYW( gamma, P, lag0, M, numLags )
%[ phi, sig2 ] = ARestYW( gamma, P, lag0, M, numLags )
%   Solve Yule-Walker equations
%
%inputs:    gamma   : Autocovariance function (must contain negative and positive lags)
%           P       : Autoregressive (AR) order
%           lag0    : Index in 'gamma' corresponding to lag 0
%           M       : Shift to be taken into account when there is a moving average part.
%           numLags : Number of lags to build the covariance matrix. Must be >= P.
%outputs:   phi     : AR coefficients
%           sig2    : Noise variance

if nargin < 4
    M = 1;
end
if nargin < 5
    numLags = P;
end

% Covariance matrix to solve YW, size numLags x P
R_x = zeros(numLags, P);
for i=1:numLags
    R_x(i,:) = gamma(lag0 + M + i - 2 : -1 : lag0 + M + i - 2 - P + 1);
end

r_x = gamma(lag0 + M : lag0 + M + numLags - 1); % autocovariance for lags M,...,M + numLags - 1

% compute autoregressive coefficients
phi = -pinv(R_x)*r_x;
phi = [1 phi.'];

% compute error variance assuming NO MA part
sig2 = real(phi*conj(gamma(lag0:lag0+P)));

end

