classdef EMAlgoMAP < handle
    % GEM algorithm for stereo audio source separation with MAP estimation of the mixing filters.
    %
    % S. Leglaive, R. Badeau, G. Richard, "Multichannel Audio Source Separation with Probabilistic Reverberation
    % Priors", submitted to IEEE Transactions on Audio, Speech and Language Processing, 2016.
    
    properties
        % Model parameters
        X; % truncated STFT of multichannel mixture [F x N x I]
        Ae; % convolutive mixing matrices for early reverberation [F x I x J]
        Al; % convolutive mixing matrices for late reverberation [F x I x J]
        Sigma_b; % additive noise variances [F x 1]
        W; % {J x 1} cell array containing the matrices of NMF bases for each source
        H; % {J x 1} cell array containing the matrices of NMF activations for each source
        
        % Statistics
        hat_P; % [F x N x J]
        hat_Rss; % [F x J x J]
        hat_Rxs; % [F x I x J]
        hat_Rxx; % [F x I x I]
        
        % Priors parameters
        phi; % AR parameters for late reverberation prior
        theta; % MA parameters for late reverberation prior
        Sigma_eps; % noise covariance matrix [I x I x F] for late reverberation prior
        Delta; % [I x J] matrix containing AR coefficients for early reverberation prior
        V; % [I x J] matrix containing the noise variances for early reverberation prior
        
        % Source posterior mean and covariance
        hat_s; % posterior mean
        Sigma_s_post; % posterior covariance matrix
        
        % Other
        log_posterior; % objective
    end
    
    methods
        
        function this = EMAlgoMAP(varargin)
            this.Set(varargin{:});
        end
        
        function Set(this,varargin)
            st_input = struct(varargin{:});
            
            if isfield(st_input, 'X')
                this.X = st_input.X;
            end
            
            if isfield(st_input, 'Ae')
                this.Ae = st_input.Ae;
            end
            
            if isfield(st_input, 'Al')
                this.Al = st_input.Al;
            end
            
            if isfield(st_input, 'Sigma_b')
                this.Sigma_b = st_input.Sigma_b;
            end
            
            if isfield(st_input, 'W')
                this.W = st_input.W;
            end
            
            if isfield(st_input, 'H')
                this.H = st_input.H;
            end
            
            if isfield(st_input, 'phi')
                this.phi = st_input.phi;
            end
            
            if isfield(st_input, 'theta')
                this.theta = st_input.theta;
            end
            
            if isfield(st_input, 'Sigma_eps')
                this.Sigma_eps = st_input.Sigma_eps;
            end
            
            if isfield(st_input, 'Delta')
                this.Delta = st_input.Delta;
            end
            
            if isfield(st_input, 'V')
                this.V = st_input.V;
            end
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            this.hat_Rss = zeros(F,J,J);
            this.hat_Rxs = zeros(F,I,J);
            this.hat_Rxx = zeros(F,I,I);
            this.hat_P = zeros(F,N,J);
            this.hat_s = zeros(F,N,J);
            this.Sigma_s_post = zeros(F,N,J,J);
        end
        
        function new = deep_copy(this)
            % Instantiate new object of the same class.
            new = feval(class(this));
            
            % Copy all non-hidden properties.
            p = properties(this);
            for i = 1:length(p)
                new.(p{i}) = this.(p{i});
            end
        end
        
        function new = copy_properties(this, new)
            % Copy all non-hidden properties.
            p = properties(this);
            for i = 1:length(p)
                new.(p{i}) = this.(p{i});
            end
        end
        
        function [] = EStep(this, b_fast)
            % b_fast: boolean indicating fast computation (b_fast = 1, default) but with potential numerical
            % instabilities, or slower computation (b_fast = 0) but without numerical instabilities.
            
            if nargin < 2
                b_fast = 1;
            end
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            
            A = this.Ae + this.Al;
            
            vecOfOnes = ones(1,N);
            
            Sigma_s = zeros(F,N,J);
            Sigma_x = zeros(F,N,2,2);
            Inv_Sigma_x = zeros(F,N,2,2);
            Gs = zeros(F,N,J,2);
            hat_Rss_fn = zeros(F,N,J,J);
            
            fprintf('   E-step\n');
            
            % compute a priori source variances
            for j = 1:J
                Sigma_s(:,:,j) = this.W{j}*this.H{j};
            end
            
            % compute Sigma_x
            Sigma_x(:,:,1,1) = this.Sigma_b * vecOfOnes;
            Sigma_x(:,:,1,2) = 0;
            Sigma_x(:,:,2,1) = 0;
            Sigma_x(:,:,2,2) = this.Sigma_b * vecOfOnes;
            for j = 1:J
                Sigma_x(:,:,1,1) = Sigma_x(:,:,1,1) + (abs(A(:,1,j)).^2 * vecOfOnes) .* Sigma_s(:,:,j);
                Sigma_x(:,:,1,2) = Sigma_x(:,:,1,2) + ((A(:,1,j) .* conj(A(:,2,j))) * vecOfOnes) .* Sigma_s(:,:,j);
                Sigma_x(:,:,2,1) = conj(Sigma_x(:,:,1,2));
                Sigma_x(:,:,2,2) = Sigma_x(:,:,2,2) + (abs(A(:,2,j)).^2 * vecOfOnes) .* Sigma_s(:,:,j);
            end

            % compute the inverse of Sigma_x
            Det_Sigma_x = Sigma_x(:,:,1,1) .* Sigma_x(:,:,2,2) - abs(Sigma_x(:,:,1,2)).^2;
            
            Inv_Sigma_x(:,:,1,1) =   Sigma_x(:,:,2,2) ./ Det_Sigma_x;
            Inv_Sigma_x(:,:,1,2) = - Sigma_x(:,:,1,2) ./ Det_Sigma_x;
            Inv_Sigma_x(:,:,2,1) =   conj(Inv_Sigma_x(:,:,1,2));
            Inv_Sigma_x(:,:,2,2) =   Sigma_x(:,:,1,1) ./ Det_Sigma_x;
            
            % compute log-posterior
            ComputeLogPosterior(this, Det_Sigma_x, Inv_Sigma_x)
            
            for j = 1:J
                % compute Wiener gain Gs
                Gs(:,:,j,1) = ((conj(A(:,1,j)) * vecOfOnes) .* Inv_Sigma_x(:,:,1,1) + ...
                    (conj(A(:,2,j)) * vecOfOnes) .* Inv_Sigma_x(:,:,2,1)) .* Sigma_s(:,:,j);
                Gs(:,:,j,2) = ((conj(A(:,1,j)) * vecOfOnes) .* Inv_Sigma_x(:,:,1,2) + ...
                    (conj(A(:,2,j)) * vecOfOnes) .* Inv_Sigma_x(:,:,2,2)) .* Sigma_s(:,:,j);
                
                % compute a posteriori source mean
                this.hat_s(:,:,j) = Gs(:,:,j,1) .* this.X(:,:,1) + Gs(:,:,j,2) .* this.X(:,:,2);
            end

            if b_fast % fast version but with potential numerical instabilities
                
                b_negVar = 0;
                
                for j1 = 1:J
                    for j2 = 1:J
                        
                        % compute a posteriori source covariance matrix
                        if j1 == j2
                            this.Sigma_s_post(:,:,j1,j2) = real( ( (j1==j2) - (Gs(:,:,j1,1) .* (A(:,1,j2) * vecOfOnes) + ...
                                Gs(:,:,j1,2) .* (A(:,2,j2) * vecOfOnes)) ) .* Sigma_s(:,:,j2) );
                        else
                            this.Sigma_s_post(:,:,j1,j2) = ( (j1==j2) - (Gs(:,:,j1,1) .* (A(:,1,j2) * vecOfOnes) + ...
                                Gs(:,:,j1,2) .* (A(:,2,j2) * vecOfOnes)) ) .* Sigma_s(:,:,j2);
                        end
                        
                        % compute Rss
                        hat_Rss_fn(:,:,j1,j2) = this.hat_s(:,:,j1) .* conj(this.hat_s(:,:,j2)) + this.Sigma_s_post(:,:,j1,j2);
                        
                    end
                    
                    this.hat_P(:,:,j1) = hat_Rss_fn(:,:,j1,j1);
                    
                    if sum(sum(real(this.Sigma_s_post(:,:,j1,j1))<0)) % if negative variances
                        b_negVar = 1;
                    end
                end
                
                if b_negVar
                    
                    % look for negative values
                    [ind1,ind2,ind3,ind4] = ind2sub(size(this.Sigma_s_post),find(real(this.Sigma_s_post) < 0));
                    
                    % look for negative values only on diagonal elements of Sigma_post_fn
                    indDiag = find(ind3 == ind4);
                    
                    % update indices
                    ind_f = ind1(indDiag);
                    ind_n = ind2(indDiag);
                    
                    if ~isempty(ind_f)
                        for ii=1:length(ind_f)
                            f = ind_f(ii);
                            n = ind_n(ii);
                            
                            % expression ensuring positive semi-definite a posteriori source covariance matrix
                            this.Sigma_s_post(f,n,:,:) = inv(diag(1./squeeze(Sigma_s(f,n,:))) + ...
                                squeeze(A(f,:,:))'*(1./this.Sigma_b(f)*eye(I))*squeeze(A(f,:,:)));
                            for j=1:J
                                this.Sigma_s_post(f,n,j,j) = real(this.Sigma_s_post(f,n,j,j));
                            end
                        end
                        
                        for j1 = 1:J
                            for j2 = 1:J
                                % compute Rss
                                hat_Rss_fn(:,:,j1,j2) = this.hat_s(:,:,j1) .* conj(this.hat_s(:,:,j2)) + this.Sigma_s_post(:,:,j1,j2);
                            end
                            % compute a posteriori source power spectrogram hat_P
                            this.hat_P(:,:,j1) = hat_Rss_fn(:,:,j1,j1);
                        end
                    end
                end
                
            else % slower version but without numerical instabilities
                % efficient computation of Sigma_s_post inspired from the code distributed with: 
                % Kounades-Bastian, D., Girin, L., Alameda-Pineda, X., Gannot, S., & Horaud, R.,
                % "A variational EM algorithm for the separation of moving sound sources," 
                % in Proc. of WASPAA 2015, available at https://team.inria.fr/perception/research/vemove/
                
                % {F x N} x [I x I] convert the [F x 1] noise variances vector to a cell array
                Sigma_b_cell = repmat( num2cell(this.Sigma_b) , 1 , N );
                
                % {F x N} x [I x J] convert the [F x I x J] mixing matrix to a cell array
                A_cell = repmat( permute( num2cell( permute(A,[2 3 1]) , [1 2] ) , [3 1 2] ) , 1 , N);
                
                inv_Sigma_s = zeros(F,N,J);
                for f=1:F
                    for n=1:N
                        inv_Sigma_s(f,n,:) = 1./Sigma_s(f,n,:);
                    end
                end
                
                % {F x N} x [J x J]
                AA = cellfun(@(A) A'*A, A_cell,'uniformoutput',false);
                
                % {F x N} x [J x J] a posteriori source covariance matrix
                Sigma_s_post_cell = cellfun(@(inv_Sigma_s,AA,Sigma_b_cell) inv( diag(inv_Sigma_s(:)) + AA / Sigma_b_cell ), num2cell(inv_Sigma_s,3), AA, Sigma_b_cell, 'uniformoutput', false);
                this.Sigma_s_post = permute(cell2mat(permute(Sigma_s_post_cell ,[4 3 1 2]) ), [3 4 1 2]);             
                
                for j1 = 1:J
                    for j2 = 1:J
                        % compute Rss
                        hat_Rss_fn(:,:,j1,j2) = this.hat_s(:,:,j1) .* conj(this.hat_s(:,:,j2)) + this.Sigma_s_post(:,:,j1,j2);
                    end
                    % compute a posteriori source power spectrogram hat_P
                    this.hat_P(:,:,j1) = hat_Rss_fn(:,:,j1,j1); 
                end
            end

            % compute Rss averaged on frames
            this.hat_Rss = squeeze(mean(hat_Rss_fn, 2));
            
            % to assure that Rss = Rss'
            for f = 1:F
                this.hat_Rss(f,:,:) = (squeeze(this.hat_Rss(f,:,:)) + squeeze(this.hat_Rss(f,:,:))') / 2;
            end
            
            % compute average Rxx
            this.hat_Rxx(:,1,1) = mean(abs(this.X(:,:,1)).^2, 2);
            this.hat_Rxx(:,2,2) = mean(abs(this.X(:,:,2)).^2, 2);
            this.hat_Rxx(:,1,2) = mean(this.X(:,:,1) .* conj(this.X(:,:,2)), 2);
            this.hat_Rxx(:,2,1) = conj(this.hat_Rxx(:,1,2));
            
            % compute average Rxs
            for j = 1:J
                this.hat_Rxs(:,1,j) = mean(this.X(:,:,1) .* conj(this.hat_s(:,:,j)), 2);
                this.hat_Rxs(:,2,j) = mean(this.X(:,:,2) .* conj(this.hat_s(:,:,j)), 2);
            end

        end
        
        function [] = MStep(this)
            fprintf('   M-step\n');
            this.PermuteParam();
            this.UpdateAe(2);
            this.UpdateDelta();
            this.UpdateAl(20);
            this.UpdateSigmab();
            this.UpdateNMF();
            this.UnpermuteParam();
        end
                                 
        function [] = UpdateAe(this, niter)
            
            if nargin < 2
                niter = 2;
            end

            [F, N, I] = size(this.X);
            J = numel(this.W);
            
            for iter=1:niter
                
                % f=1
                vecAtmp = N*this.hat_Rxs(:,:,1) - N*this.Al(:,:,1)*this.hat_Rss(:,:,1) + ...
                    (this.Sigma_b(1)*eye(I))*(this.V.*(conj(this.Delta).*this.Ae(:,:,2)));
                
                vecAe = (N*kron(this.hat_Rss(:,:,1).',eye(I)) + kron(eye(J),this.Sigma_b(1)*eye(I)).*(ones(I*J,1)*this.V(:).') )\vecAtmp(:);
                this.Ae(:,:,1) = reshape(vecAe,I,J);
                
                % f=2,...,F-1
                for f=2:F-1
                    vecAtmp = N*this.hat_Rxs(:,:,f) - N*this.Al(:,:,f)*this.hat_Rss(:,:,f) + ...
                        (this.Sigma_b(f)*eye(I))*(this.V.*( conj(this.Delta).*this.Ae(:,:,f+1) + this.Delta.*this.Ae(:,:,f-1) ));
                    
                    vecAe = (N*kron(this.hat_Rss(:,:,f).',eye(I)) + kron(eye(J),this.Sigma_b(f)*eye(I)).*(ones(I*J,1)*2*this.V(:).') )\vecAtmp(:);
                    this.Ae(:,:,f) = reshape(vecAe,I,J);
                end
                
                % f=F
                vecAtmp = N*this.hat_Rxs(:,:,F) - N*this.Al(:,:,F)*this.hat_Rss(:,:,F) + ...
                    (this.Sigma_b(F)*eye(I))*(this.V.*( this.Delta.*this.Ae(:,:,F-1) ) );
                
                vecAe = (N*kron(this.hat_Rss(:,:,F).',eye(I)) + kron(eye(J),this.Sigma_b(F)*eye(I)).*(ones(I*J,1)*2*this.V(:).') )\vecAtmp(:);
                this.Ae(:,:,F) = reshape(vecAe,I,J);
                
            end
            
        end
        
        function [] = UpdateDelta(this)
            
            this.Delta = sum(this.Ae(:,:,2:end).*conj(this.Ae(:,:,1:end-1)),3);
            this.Delta = this.Delta./abs(this.Delta);
            
        end
                
        function [] = UpdateAl(this, niter)
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            
            % compute the inverse of Sigma_eps matrix
            Det_Sigma_eps = this.Sigma_eps(1,1,:) .* this.Sigma_eps(2,2,:) - abs(this.Sigma_eps(1,2,:)).^2;
            Inv_Sigma_eps(1,1,:) =   this.Sigma_eps(2,2,:) ./ Det_Sigma_eps;
            Inv_Sigma_eps(1,2,:) = - this.Sigma_eps(1,2,:) ./ Det_Sigma_eps;
            Inv_Sigma_eps(2,1,:) =   conj(Inv_Sigma_eps(1,2,:));
            Inv_Sigma_eps(2,2,:) =   this.Sigma_eps(1,1,:) ./ Det_Sigma_eps;
            
            % compute the inverse of Sigma_b matrix
            inv_Sigmab = zeros(I,I,F);
            inv_Sigmab(1,1,:) = 1./this.Sigma_b;
            inv_Sigmab(2,2,:) = 1./this.Sigma_b;
            
            % compute static part of the preconditioning matrix assuming diagonal and stationary Sigma_eps
            sigma_eps2 = this.Sigma_eps(1,1,1);
            psi =  computePsiFromARMA( this.phi, this.theta, 2*F-2);
            d_stat = sum(abs(psi).^2)/sigma_eps2*ones(I*J,1);
            
            % compute vec(Al_f)
            vec_Al = reshape(this.Al, I*J, F);
            
            % compute gradient (1)
            g1 = zeros(I*J,F);
            
            Inv_Sigma_eps_ext = zeros(I*J,I*J,F);
            for j=1:J
                Inv_Sigma_eps_ext(I*(j-1)+1:j*I,I*(j-1)+1:j*I, :) = Inv_Sigma_eps;
            end
            
            vec_epsilon = filter(this.phi, this.theta, vec_Al, [], 2);
            for i=1:I*J
                g1(i,:) = sum(squeeze(Inv_Sigma_eps_ext(i,:,:)).*vec_epsilon, 1);
            end
            g1 = filter(conj(this.phi), conj(this.theta), g1(:,end:-1:1), [], 2);
            g1 = g1(:,end:-1:1);
            
            % compute gradient (2)
            g2 = zeros(I*J,F);
            Rss_x_invSigmab = zeros(I*J, I*J, F);
            for j1=1:J
                for j2=1:J
                    Rss_x_invSigmab(I*(j1-1)+1:j1*I,I*(j2-1)+1:j2*I, :) = repmat(this.hat_Rss(j2, j1, :), [I,I,1]) .* inv_Sigmab;
                end
            end
            for i=1:I*J
                g2(i,:) = N*sum(squeeze(Rss_x_invSigmab(i,:,:)).*vec_Al, 1);
            end
            
            % compute gradient (3)
            Ae_x_Rss = zeros(I,J,F);
            for j1=1:J
                for j2=1:J
                    Ae_x_Rss(1,j1,:) = Ae_x_Rss(1,j1,:) + this.Ae(1,j2,:).*this.hat_Rss(j2,j1,:);
                    Ae_x_Rss(2,j1,:) = Ae_x_Rss(2,j1,:) + this.Ae(2,j2,:).*this.hat_Rss(j2,j1,:);
                end
            end
            tmp = Ae_x_Rss - this.hat_Rxs;
            G3 = zeros(I,J,F);
            for j=1:J
                G3(1,j,:)  = inv_Sigmab(1,1,:).*tmp(1,j,:) + inv_Sigmab(1,2,:).*tmp(2,j,:);
                G3(2,j,:)  = inv_Sigmab(2,1,:).*tmp(1,j,:) + inv_Sigmab(2,2,:).*tmp(2,j,:);
            end
            g3 = N*reshape(G3, I*J, F);
            
            % compute gradient (final)
            g = g1 + g2 + g3;
            
            % preconditionning
            D = zeros(I*J, F);
            for ij=1:I*J
                D(ij,:) = squeeze(Rss_x_invSigmab(ij,ij,:));
            end
            d = N*D + d_stat*ones(1,F);
            d = d(:);
            inv_d = 1./d;
            
            % descent direction
            w = diag(inv_d)*g(:);
            w = reshape(w,I*J,F);
            
            iter = 1;
        
            while iter <= niter 
                
                % compute gamma (1)
                gamma1 = zeros(I*J,F);
                for i=1:I*J
                    gamma1(i,:) = N*sum(squeeze(Rss_x_invSigmab(i,:,:)).*w, 1);
                end
                
                % compute gamma (2)
                gamma2 = zeros(I*J,F);
                tmp = filter(this.phi, this.theta, w, [], 2);
                for i=1:I*J
                    gamma2(i,:) = sum(squeeze(Inv_Sigma_eps_ext(i,:,:)).*tmp, 1);
                end
                gamma2 = filter(conj(this.phi), conj(this.theta), gamma2(:,end:-1:1), [], 2);
                gamma2 = gamma2(:,end:-1:1);
                
                % compute gamma (final)
                gamma = gamma1 + gamma2;
                
                % compute mu
                mu = w(:)'*g(:)/(w(:)'*gamma(:));
                mu = real(mu);                
                
                % update Al_f
                vec_Al = vec_Al - mu*w;
                this.Al = reshape(vec_Al,I,J,F);
                
                % update gradient (1)
                g1 = zeros(I*J,F);
                vec_epsilon = filter(this.phi, this.theta, vec_Al, [], 2);
                for i=1:I*J
                    g1(i,:) = sum(squeeze(Inv_Sigma_eps_ext(i,:,:)).*vec_epsilon, 1);
                end
                g1 = filter(conj(this.phi), conj(this.theta), g1(:,end:-1:1), [], 2);
                g1 = g1(:,end:-1:1);
                
                % update gradient (2)
                g2 = zeros(I*J,F);
                for i=1:I*J
                    g2(i,:) = N*sum(squeeze(Rss_x_invSigmab(i,:,:)).*vec_Al, 1);
                end
                
                % update gradient (final)
                g = g1 + g2 + g3;
                
                % preconditionning
                g_pc = diag(inv_d)*g(:);
                g_pc = reshape(g_pc,I*J,F);
                
                % alpha
                alpha = - gamma(:)'*g_pc(:)/(w(:)'*gamma(:));
                alpha = real(alpha);
                
                % update descent direction
                w = g_pc + alpha*w;
                
                iter = iter + 1;
                
            end
        end
        
        function [] = UpdateSigmab(this)
            [F, N, I] = size(this.X);
            J = numel(this.W);
            A = this.Ae + this.Al;
            
            for f = 1:F
                this.Sigma_b(f) = 0.5 * real(trace(this.hat_Rxx(:,:,f) - ...
                    A(:,:,f) * this.hat_Rxs(:,:,f)' - ...
                    this.hat_Rxs(:,:,f) * A(:,:,f)' + ...
                    A(:,:,f) * this.hat_Rss(:,:,f) * A(:,:,f)'));
            end

            if sum(this.Sigma_b < 0) % if small negative variances (should not happen if b_fast = 0 at E-step)
                tmp = zeros(N,1);
                ind_f = find(this.Sigma_b<0);
                for f=1:length(ind_f)
                    for n=1:N
                        C = squeeze(this.Sigma_s_post(ind_f(f),n,:,:));
                        [V,D] = eig(C);
                        if sum(real(diag(D))<0)
                            [~,ind_neg_EV] = min(real(diag(D)));
                            V1 = V(:,ind_neg_EV);
                            C = C + V1*V1'*real(eps(D(ind_neg_EV,ind_neg_EV))-D(ind_neg_EV,ind_neg_EV) + 1e-10);
                        end
                        tmp(n) = trace(A(:,:,ind_f(f))*C*A(:,:,ind_f(f))') + ...
                            (squeeze(this.X(ind_f(f),n,:)) - A(:,:,ind_f(f))*squeeze(this.hat_s(ind_f(f),n,:)))'*...
                            (squeeze(this.X(ind_f(f),n,:)) - A(:,:,ind_f(f))*squeeze(this.hat_s(ind_f(f),n,:)));
                    end
                    this.Sigma_b(ind_f(f)) = 0.5*mean(real(tmp));
                end
            end
        end
       
        function [] = UpdateNMF(this)
            [F, N, I] = size(this.X);
            J = numel(this.W);
            for j=1:J
                [this.W{j}, this.H{j}, ~] = this.ISNMF(real(this.hat_P(:,:,j)) + eps, this.W{j}, this.H{j}, 10);
            end
        end

        function [W, H, E] = ISNMF(this, V, W0, H0, iter_num)
            
            if nargin < 4 || isempty(iter_num)
                iter_num = 100;
            end;
            
            [F,N] = size(V);
            
            W = W0;
            H = H0;
            
            S = sum(V(:));
            matOfOnes = ones(F,N);
            WH = W*H + eps;
            
            E = zeros(1, iter_num);
            
            for iter = 1:iter_num
                % update of W
                W = W .* (( V./((WH).^2) )*H' + eps) ./ ((1./(WH))*H' + eps);
                
                % normalization
                w = sum(W,1);
                d = diag(1./w);
                W = W*d; % Normalizatiosn of columns of W
                d = diag(w);
                H = d*H; % Energy transfer to H
                
                WH = W*H + eps;
                
                % update of H
                H = H .* (W'*(V./((WH).^2)) + eps) ./ (W'*(1./(WH)) + eps);
                WH = W*H + eps;
                
                % Error with IS divergence:
                E(iter) = sum(sum(V./(WH + eps) - log(V./(WH + eps) + eps) - matOfOnes))/S; % normalized error

            end
        end
        
        function [] = ComputeLogPosterior(this, Det_Sigma_x, Inv_Sigma_x)
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            
            log_like = -sum(sum(log(Det_Sigma_x * (pi^I)) + ...
                Inv_Sigma_x(:,:,1,1) .* (abs(this.X(:,:,1)).^2) + ...
                Inv_Sigma_x(:,:,2,2) .* (abs(this.X(:,:,2)).^2) + ...
                2 * real(Inv_Sigma_x(:,:,1,2) .* conj(this.X(:,:,1)) .* this.X(:,:,2)) ));
            
            log_prior_early = ComputeLogPriorEarly(this);
            
            log_prior_late = ComputeLogPriorLate(this);
            
            this.log_posterior = real(log_like + log_prior_early + log_prior_late);
            
        end
        
        function [Q_fun] = ComputeQFunction(this)
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            
            if size(this.Ae,1) ~= F
                b_permute = 1;
                this.UnpermuteParam();
            else
                b_permute = 0;
            end
            
            A = this.Ae + this.Al;
            
            expect_complete_data_log_like = 0;
            
            for f=1:F
                % compute contribution of complete data log-likelihood
                inv_Sigma_b_f = inv(eye(I)*this.Sigma_b(f));
                expect_complete_data_log_like = expect_complete_data_log_like - N*( log((pi*this.Sigma_b(f)).^I) + trace( ...
                    inv_Sigma_b_f*squeeze(this.hat_Rxx(f,:,:)) - ...
                    inv_Sigma_b_f*squeeze(A(f,:,:))*squeeze(this.hat_Rxs(f,:,:))' - ...
                    inv_Sigma_b_f*squeeze(this.hat_Rxs(f,:,:))*squeeze(A(f,:,:))' + ...
                    inv_Sigma_b_f*squeeze(A(f,:,:))*squeeze(this.hat_Rss(f,:,:))*squeeze(A(f,:,:))'));
            end
            
            for j=1:J
                V_j = this.W{j}*this.H{j};
                expect_complete_data_log_like = expect_complete_data_log_like - sum(sum(log(V_j + eps) + this.hat_P(:,:,j)./(V_j + eps)));
            end
            
            [log_prior_early] = ComputeLogPriorEarly(this);
            [log_prior_late] = ComputeLogPriorLate(this);
            
            Q_fun = real(expect_complete_data_log_like + log_prior_early + log_prior_late);
            
            if b_permute == 1
                this.PermuteParam();
            end
            
        end
        
        function [log_prior_early] = ComputeLogPriorEarly(this)
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            log_prior_early = -(F-1)*sum(log(pi./this.V(:)));
            
            for f=2:F
                log_prior_early = log_prior_early - trace( (this.V.*squeeze(this.Ae(f,:,:)) - this.V.*this.Delta.*squeeze(this.Ae(f-1,:,:)))*squeeze(this.Ae(f,:,:))' + ...
                    (this.V.*squeeze(this.Ae(f-1,:,:)) - this.V.*conj(this.Delta).*squeeze(this.Ae(f,:,:)))*squeeze(this.Ae(f-1,:,:))' );
            end
        end
        
        function [log_prior_late] = ComputeLogPriorLate(this)
            
            [F, N, I] = size(this.X);
            J = numel(this.W);
            this.Sigma_eps = permute(this.Sigma_eps, [3,1,2]);
            
            % compute the inverse of Sigma_eps matrix
            Det_Sigma_eps = this.Sigma_eps(:,1,1) .* this.Sigma_eps(:,2,2) - abs(this.Sigma_eps(:,1,2)).^2;
            Inv_Sigma_eps(:,1,1) =   this.Sigma_eps(:,2,2) ./ Det_Sigma_eps;
            Inv_Sigma_eps(:,1,2) = - this.Sigma_eps(:,1,2) ./ Det_Sigma_eps;
            Inv_Sigma_eps(:,2,1) =   conj(Inv_Sigma_eps(:,1,2));
            Inv_Sigma_eps(:,2,2) =   this.Sigma_eps(:,1,1) ./ Det_Sigma_eps;
            
            log_prior_late = 0;
            
            for f=1:F
                epsilon = filter(this.phi, this.theta, this.Al, [], 1);
                log_prior_late = log_prior_late - J*log(det(pi*squeeze(this.Sigma_eps(f,:,:))))  - ...
                    trace(squeeze(epsilon(f,:,:))'*squeeze(Inv_Sigma_eps(f,:,:))*squeeze(epsilon(f,:,:)));
            end
            
            this.Sigma_eps = permute(this.Sigma_eps, [2,3,1]);
        end
        
        function [] = PermuteParam(this)
            this.Ae = permute(this.Ae, [2,3,1]);
            this.Al = permute(this.Al, [2,3,1]);
            this.hat_Rxs = permute(this.hat_Rxs, [2,3,1]);
            this.hat_Rss = permute(this.hat_Rss, [2,3,1]);
            this.hat_Rxx = permute(this.hat_Rxx, [2,3,1]);
        end
        
        function [] = UnpermuteParam(this)
            this.Ae = permute(this.Ae, [3,1,2]);
            this.Al = permute(this.Al, [3,1,2]);
            this.hat_Rxs = permute(this.hat_Rxs, [3,1,2]);
            this.hat_Rss = permute(this.hat_Rss, [3,1,2]);
            this.hat_Rxx = permute(this.hat_Rxx, [3,1,2]);
        end

    end
    
end

