function [phi, theta] = computeARMAParameters(Fs, T60, room_size, T, P, Q)
%[phi, theta, sigma2_eps, sigma2_rev] = computeARMAParameters(Fs, T60, room_size, T, P, Q)
%   Compute the ARMA(P,Q) parametrization of the theoretical autocovariance function for the frequency response of
%   late reverberation.
%inputs:    Fs          : sampling rate
%           T60         : reverberation time
%           room_size   : dimensions of the room
%           T           : RIR length
%           P           : AR order
%           Q           : MA order
%outputs:   phi         : AR coefficients
%           theta       : MA coefficients
%           
% For more details see S. Leglaive, R. Badeau, G. Richard, "Autoregressive moving average modeling of late 
% reverberation in the frequency domain", in Proc. of the European Signal Processing Conference (EUSIPCO),
% Budapest, Hungary, 2016.

c=343; % sound velocity
% room dimensions
Lx = room_size(1); 
Ly = room_size(2);
Lz = room_size(3);
V = Lx*Ly*Lz; % volume of the room

%% theoretical variance

Sxz=Lx*Lz; 
Syz=Ly*Lz; 
Sxy=Lx*Ly; 
S=2*(Sxz+Syz+Sxy); % Total area of shoebox room surfaces
beta_abs = 1 - exp(-24*log(10)*V/(c*S*T60)); % absorption coefficient
sigma2_rev = 75*(1-beta_abs)/(pi*beta_abs*S);

%% thetoretical ACVF

t0 = 2*sqrt(V)*1e-3; % time instant between early and late parts of RIR

f=(0:T-1)';
K = T/2;
tau = T60/(3*log(10))*Fs; % decay parameter
P02 = sigma2_rev*exp(2*T/tau)*(1-exp(2/tau))/(1 - exp(2*(T-floor(t0*Fs) +1)/tau));
gamma_theo = P02*exp(-2*T/tau)*(1 - exp((1i*2*pi*f/T + 2/tau)*(T-floor(t0*Fs)+1)))./( 1 -exp((1i*2*pi*f/T + 2/tau)));
gamma_theo = circshift(gamma_theo,K-1);

%% Compute AR parameters

% if Q==0
%     M = 1;
% else
%     M = max(P,Q+1);
% end

M = Q+1;

% compute AR parameters
[ phi, ~ ] = ARestYW( gamma_theo, P, K, M, 2*P );

%% remove AR part

if Q==0
    return;
end

gamma_theo_woAR = filter(phi,1,gamma_theo);
gamma_theo_woAR = filter(conj(phi),1,gamma_theo_woAR(end:-1:1));
gamma_theo_woAR = gamma_theo_woAR(end:-1:1);
gamma_MA = gamma_theo_woAR;

%% compute MA parameters

% Approximate MA by a high-order AR
P_high = 10*Q;
[phi_high, sigma2_high] = ARestYW( gamma_MA, P_high, K);

% Compute autocovariance function of the sequence phi_high
% y = conv(phi1,conj(phi1(end:-1:1)))/length(phi1);
r_phi = xcorr(phi_high, 'biased');
r_phi = r_phi(:);

% fit new AR to get MA coeffs~
[theta, ~] = ARestYW( r_phi, Q, P_high + 1 );

%% compute the variance of the innovation

% compute linear process representation
% we have to identify : (1 + phi_1*z + phi_2*z^2 + ... + phi_P*z^P)*(psi_0 + psi_1*z + psi_2*z^2 + ...) =
%                       (1 + theta_1*z + theta_2*z^2 + ... + theta_Q*
psi = zeros(Q+1,1);
psi(1) = 1;

for j=1:Q
    for k=1:P
        if j-k>=0
            psi(j+1) = psi(j+1) - phi(k+1)*psi(j-k+1);
        end
    end
    psi(j+1) = psi(j+1) + theta(j+1);
end

sigma2_eps = real((phi*gamma_theo(K:-1:K-P))/(theta*conj(psi)));

% sigma2_eps = sigma2_high;

