% Initialization procedure. Different runs lead to different initializations.

%% General parameters

addpath('./aux_tools');

dataDir = './sounds'; % directory of true source images and mix 
resultsDir = dataDir; % directory to write estimated source images

I = 2; % number of microphones
J = 4; % number of sources
Kj = 4; % number of clusters per source
K = J*Kj; % total number of clusters

%% Compute STFT of the mix
fprintf('Initialize parameters\n');

% get mix wave file name
wavFile = fullfile(dataDir,'mix.wav');

% read wav file
[x, fs] = wavread(wavFile);
x = x.';
mix_nsamp = size(x,2);

% length of STFT analysis window
stft_win_len = 2048;

% compute STFT
X = stft_multi(x,stft_win_len);

F = size(X,1);
N = size(X,2);

%% Compute NMF on stacked left and right channels

% In this block the code is coming from Init_KL_NMF_fr_sep_sources.m distributed with [Ozerov and Fevotte, 2010], but
% we use is-nmf

% stack left and right channels
X2 = [squeeze(X(:,:,1)) ; squeeze(X(:,:,2))];

% compute power spectrogram
P = abs(X2).^2;

% some constants
small_noise_var = 3e-11;       % small noise variance

V = P + randn(2*F, N).^2 * small_noise_var;

% run multiplicative algorithm with 10 iterations and 5 random
% initializations, and keep the best initialization
for init_ind = 1:5
    W_src = 0.3 + 1.7 * rand(2*F, K) * sqrt(mean(V(:)));
    H_src = 0.3 + 1.7 * rand(K, N) * sqrt(mean(V(:)));
    
    [W_src,H_src,E] = nmf_is(V,W_src,H_src,10);
    
    E_end = E(end);
    
    if init_ind > 1
        if E_end < E_end_best
            W_src_best = W_src;
            H_src_best = H_src;
            E_end_best = E_end;
        end;
    else
        W_src_best = W_src;
        H_src_best = H_src;
        E_end_best = E_end;
    end;
end;

% now run 50 iterations of multiplicative algorithm with best initialization
[W, H] = nmf_is(V,W_src_best,H_src_best,50);

W = max(W, small_noise_var * 100);
H = max(H, small_noise_var * 10); 

%% Reconstruct K components spectrogram matrices by Wiener filtering

C = zeros(K,2*F,N);

for f=1:2*F
    for n=1:N
        den = W(f,:)*H(:,n);
        C(:,f,n) = W(f,:).'.*H(:,n)/den*X2(f,n);
    end
end

CL = C(:,1:F,:);
CR = C(:,F+1:end,:);

%% Construct K ad-hoc mixing filters for left and right channels

CL_tilde = zeros(size(CL));
CR_tilde = zeros(size(CR));

for k=1:K
    phi = exp(1j*angle(squeeze(CL(k,:,:))));
    CL_tilde(k,:,:) = squeeze(CL(k,:,:))./(phi+eps);
    CR_tilde(k,:,:) = squeeze(CR(k,:,:))./(phi+eps);
end

CL_bar = mean(CL_tilde,3);
CL_bar = permute(CL_bar,[2 1]);
CR_bar = mean(CR_tilde,3);
CR_bar = permute(CR_bar,[2 1]);


%% Cluster the mixing filters according to the TDoA

data = unwrap(angle(CL_bar./CR_bar));
[idx,clust] = kmeans(data.', J);

%% Compute final mixing filters

A_init = zeros(2,J,F);

for j=1:J
    A_init(1,j,:) = mean(CL_bar(:,idx==j),2);
    A_init(2,j,:) = mean(CR_bar(:,idx==j),2);
end

%% Binary masking to compute separated sources from estimated mixing filters

S_bm = sep_binmask(X,A_init);

% Computation of the spatial source images
Ie_bm=src_image(S_bm,A_init);
ie_bm=istft_multi(Ie_bm,mix_nsamp);
for j=1:J
    wavwrite(reshape(ie_bm(j,:,:),mix_nsamp,2),fs,fullfile(resultsDir, ['sim_init_adHoc' int2str(j) '.wav']));
end

%% Evaluation

% % Evaluation of the estimated source images (binary masking used for initialization of EM)
% fprintf('Evaluation of the estimated source images:\n');
% i=zeros(J,mix_nsamp,2);
% for j=1:J
%     i(j,:,:)=reshape(wavread(fullfile(dataDir, ['sim' int2str(j) '.wav'])),1,mix_nsamp,2);
% end
% [SDRinit,ISRinit,SIRinit,SARinit,perminit]=bss_eval_images(ie_bm,i);

%% Compute KL-NMF on separeted sources

% compute ISTFT and STFT
Se_bm_rec = stft_multi(istft_multi(S_bm,mix_nsamp),stft_win_len);

% initialize W and H from separated sources
[~, ~, ~, W_init, H_init, E] = Init_KL_NMF_fr_sep_sources(Se_bm_rec, 10);
        
%% Noise variance initialization

mix_psd = 0.5 * (mean(abs(X(:,:,1)).^2 + abs(X(:,:,2)).^2, 2));
Sigma_b_init = mix_psd / 100;

%% Save parameters

save('init_param.mat','A_init', 'W_init', 'H_init', 'Sigma_b_init');
