clear all
close all

addpath('./aux_tools')

%% Global parameters

dataDir = './sounds'; % directory of true source images and mix 
resultsDir = dataDir; % directory to write estimated source images
resultsWithPriorPrefix = 'sim_EMPrior_';

J = 4; % number of sources
I = 2; % number of microphones
niter = 500; % number of iterations for the GEM algorithm

% room dimensions
Lx = 4.45;
Ly = 3.55;
Lz = 2.5;
room_size = [Lx; Ly; Lz];

T60 = 0.1281; % reverberation time

Vol = prod(room_size); % volume of the room
t0 = sqrt(Vol)*2e-3; % mixing time
   
%% Compute STFT

% get mix wave file name
wavFile = fullfile(dataDir,'mix.wav');

% read wav file and compute time-frequency representation
fprintf('Compute STFT\n');

[x, fs]=wavread(wavFile);
x = x.';
mix_nsamp = size(x,2);

stft_win_len = 2048;
X = stft_multi(x,stft_win_len);

F = size(X,1);
N = size(X,2);

%% Initialization

if ~exist('init_param.mat','file')
    init_param;
end
load('init_param.mat');

% Split A_init into early and late parts
A_init_all = cat(3,A_init,conj(A_init(:,:,end-1:-1:2)));
Ae_init = zeros(size(A_init));
Al_init = zeros(size(A_init));

for i=1:I
    for j=1:J
        a_init = real(ifft(A_init_all(i,j,:)));   
        tmp = squeeze(a_init);
        tmp(floor(t0*fs)+1:end) = 0;
        tmp = fft(tmp);
        Ae_init(i,j,:) = tmp(1:end/2+1);
        Ae_init(i,j,:) = Ae_init(i,j,:) - mean(Ae_init(i,j,:));
        
        tmp = squeeze(a_init);
        tmp(1:round(t0*fs)) = 0;
        tmp = fft(tmp);
        Al_init(i,j,:) = tmp(1:end/2+1);
        Al_init(i,j,:) = Al_init(i,j,:) - mean(Al_init(i,j,:));
    end
end

A_init = Ae_init + Al_init;

%% Hyperparameters for early reverberation prior

% AR parameters
Delta = sum(Ae_init(:,:,2:F).*conj(Ae_init(:,:,1:F-1)),3);
Delta = Delta./abs(Delta);

% variance
sigma2_early = 1e-7;
V = (1/sigma2_early)*ones(I,J);

%% Hyperparameters for late reverberation prior

% ARMA parameters
[phi, theta] = computeARMAParameters(fs, T60, room_size, stft_win_len, 7, 2);

% variance
sigma2_eps = 1e-7;
Sigma_eps = repmat(eye(2)* sigma2_eps,[1,1,F]);

%% Run GEM algorithm

Ae_init = permute(Ae_init, [3 1 2]);
Al_init = permute(Al_init, [3 1 2]);
A_init = permute(A_init, [3 1 2]);

ss_algoMAP = EMAlgoMAP('X', X, 'Ae', Ae_init, 'Al', Al_init, 'Sigma_b', Sigma_b_init, 'W', {W_init}, 'H', {H_init}, 'phi', phi, 'theta', theta, 'Sigma_eps', Sigma_eps, 'Delta', Delta, 'V', V);

log_posterior = zeros(niter,1);

for iter = 1:niter
    fprintf('EM iteration %d of %d\n', iter, niter);

    ss_algoMAP.EStep();

    log_posterior(iter) = ss_algoMAP.log_posterior;
    fprintf('log-posterior: %.3f\n', log_posterior(iter)/(F*N));

    ss_algoMAP.MStep;
end

ss_algoMAP.EStep;


%% Computation of the spatial source images

Ae = ss_algoMAP.Ae;
Al = ss_algoMAP.Al;
Ae = permute(Ae, [2 3 1]);
Al = permute(Al, [2 3 1]);

A_MAP = Ae + Al;
S_MAP = ss_algoMAP.hat_s;

fprintf('Computation of the spatial source images\n');

Ie_MAP=src_image(S_MAP,A_MAP);
ie_MAP=istft_multi(Ie_MAP,mix_nsamp);

for k=1:J,
    wavwrite(reshape(ie_MAP(k,:,:),mix_nsamp,2),fs,fullfile(resultsDir, [resultsWithPriorPrefix num2str(k) '.wav']));
end

%% Computation of the scores

% fprintf('Computation of the scores\n');
% i_true = zeros(J,mix_nsamp,2);
% for j=1:J
%     i_true(j,:,:)=reshape(wavread(fullfile(dataDir, ['sim' int2str(j) '.wav'])),1,mix_nsamp,2);
% end
% [SDRinit,ISRinit,SIRinit,SARinit,perminit]=bss_eval_images(ie_MAP,i_true);
